#include "ITL_Preconditioner.h"

namespace AMDiS {
  template< typename Vector >
  struct PreconPair { 
    /// Pointer to the left preconditioner
    ITL_BasePreconditioner< Vector >* l;

    /// Pointer to the right preconditioner
    ITL_BasePreconditioner< Vector >* r;
    PreconPair():
      l(NULL), r(NULL)
    {}

  };

  struct BaseITL_runner { 
    OEMSolver& oem;
    BaseITL_runner(OEMSolver* oemPtr):
      oem(*oemPtr)
    {
      TEST_EXIT(oemPtr != NULL)("need a real oemsolver\n");
    }

    template< typename Vector , typename Matrix >
    void setPrecon(PreconPair< Vector >& pair, const Matrix& matrix)
    {
      /// Creator for the left preconditioner
      CreatorInterface< ITL_BasePreconditioner< Vector > >* leftCreator(NULL);

      /// Creator for the right preconditioner
      CreatorInterface< ITL_BasePreconditioner< Vector > >* rightCreator(NULL);

      std::string preconType("no");
      Parameters::get(oem.getName() + "->left precon", preconType);

      leftCreator = CreatorMap<ITL_BasePreconditioner< Vector > >::getCreator(preconType);
      TEST_EXIT(leftCreator != NULL)
	("there is no creator for the given left preconditioner");

      preconType = "no";
      Parameters::get(oem.getName() + "->right precon", preconType);
      rightCreator = CreatorMap<ITL_BasePreconditioner< Vector > >::getCreator(preconType);
      TEST_EXIT(rightCreator != NULL)
	("there is no creator for the given right preconditioner");

      pair.l = leftCreator->create(matrix);
      pair.r = rightCreator->create(matrix);
    }
  };

  template< typename ITLSolver, typename Matrix, typename Vector >
  struct ITL_OEMSolver_para_runner : BaseITL_runner {
    ITL_OEMSolver_para_runner(OEMSolver* oem_):
      BaseITL_runner(oem_),
      ell(1)
    {
      ell=oem.getMaxIterations();
      Parameters::get(oem.getName() + "->ell", ell);
    }

    void init(const Matrix& A)
    {
	BaseITL_runner::setPrecon(preconPair, A); 
    }

    int solve(const Matrix& A, Vector& x, Vector& b) 
    {
      FUNCNAME("ITL_OEMSolver_para_runner::solve()");
      TEST_EXIT(preconPair.l != NULL)("there is no left preconditioner");
      TEST_EXIT(preconPair.r != NULL)("there is no right preconditioner");

      Vector r(A * x - b); 
      itl::cyclic_iteration<typename Vector::value_type> iter(r, oem.getMaxIterations(), oem.getRelative(), 
							      oem.getTolerance(), oem.getPrint_cycle());

      int error = solver(A, x, b, *(preconPair.l), *(preconPair.r), iter, ell);
      oem.setIterations(iter.iterations());
      oem.setResidual(iter.resid());
      oem.setErrorCode(error);

      return error;
    }
  private:
    PreconPair< Vector > preconPair;
    ITLSolver solver;
    int ell;
  };

  template< typename ITLSolver, typename Matrix, typename Vector >
  struct ITL_OEMSolver_runner : BaseITL_runner {
   
   
    ITL_OEMSolver_runner(OEMSolver* oem):
      BaseITL_runner(oem)
    {}

    void init(const Matrix& A)
    {
	BaseITL_runner::setPrecon(preconPair, A);
    }
    int solve(const Matrix& A , Vector& x, Vector& b) 
    {
      FUNCNAME("ITL_OEMSolver_runner::solve()");
      TEST_EXIT(preconPair.l != NULL)("there is no left preconditioner");
      TEST_EXIT(preconPair.r != NULL)("there is no right preconditioner");

      Vector r(A * x - b); 
      itl::cyclic_iteration<typename Matrix::value_type> iter(r, oem.getMaxIterations(), oem.getRelative(), 
							      oem.getTolerance(), oem.getPrint_cycle());

      int error = solver(A, x, b, *(preconPair.l), *(preconPair.r), iter);
      oem.setErrorCode(error);
      oem.setIterations(iter.iterations());
      oem.setResidual(iter.resid());

      return error;
    }
    private:
    ITLSolver solver;
    PreconPair< Vector > preconPair;
  };
}
