//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include <fstream>
#include <stdint.h>

#include "ArhReader.h"
#include "Mesh.h"
#include "MeshStructure.h"
#include "Traverse.h"
#include "DOFVector.h"

namespace AMDiS {
  
  using namespace std;


  void ArhReader::read(string filename, Mesh *mesh,
		       DOFVector<double>* vec0, 
		       DOFVector<double>* vec1,
		       DOFVector<double>* vec2)
  {
    vector<DOFVector<double>*> vecs(0);
    if (vec0 != NULL)
      vecs.push_back(vec0);
    if (vec1 != NULL)
      vecs.push_back(vec1);
    if (vec2 != NULL)
      vecs.push_back(vec2);

    ArhReader::read(filename, mesh, vecs);
  }


  void ArhReader::read(string filename, Mesh *mesh, 
		       vector<DOFVector<double>*> vecs)
  {
    FUNCNAME("ArhReader::read()");

    // === Get set of all macro elements in mesh. ===
    std::set<int> macroInMesh;
    for (std::deque<MacroElement*>::iterator it = mesh->getMacroElements().begin();
	 it != mesh->getMacroElements().end(); ++it)
      macroInMesh.insert((*it)->getIndex());
    

    RefinementManager *refManager = NULL;
    switch (mesh->getDim()) {
    case 2:
      refManager = new RefinementManager2d();
      break;
    case 3:
      refManager = new RefinementManager3d();
      break;
    default:
      ERROR_EXIT("Should not happen!\n");
    }

    ifstream file;
    file.open(filename.c_str(), ios::in | ios::binary);

    string typeId = "";
    uint32_t nMacroElements = 0;
    uint32_t nValueVectors = 0;
    uint32_t nAllValues = 0;

    file.read(const_cast<char*>(typeId.data()), 4);
    file.read(reinterpret_cast<char*>(&nMacroElements), 4);
    file.read(reinterpret_cast<char*>(&nValueVectors), 4);
    file.read(reinterpret_cast<char*>(&nAllValues), 4);

    TEST_EXIT(nValueVectors == vecs.size())
      ("File has %d vectors, but %d DOFVectors are provided!\n", 
       nValueVectors, vecs.size());

    for (unsigned int i = 0; i < nMacroElements; i++) {
      uint32_t elIndex = 0;
      uint32_t nStructureCodes = 0;
      uint32_t codeSize = 0;

      file.read(reinterpret_cast<char*>(&elIndex), 4);
      file.read(reinterpret_cast<char*>(&nStructureCodes), 4);
      file.read(reinterpret_cast<char*>(&codeSize), 4);

      vector<uint64_t> structureCode(nStructureCodes);
      file.read(reinterpret_cast<char*>(&(structureCode[0])), 8 * nStructureCodes);

      MeshStructure elementStructure;
      elementStructure.init(structureCode, codeSize);
      if (macroInMesh.count(elIndex) == 1)
	elementStructure.fitMeshToStructure(mesh, refManager, false, elIndex);

      uint32_t nValuesPerVector = 0;
      file.read(reinterpret_cast<char*>(&nValuesPerVector), 4);

      for (unsigned int j = 0; j < nValueVectors; j++) {
	vector<double> values(nValuesPerVector);
	file.read(reinterpret_cast<char*>(&(values[0])), 8 * nValuesPerVector);
	if (macroInMesh.count(elIndex) == 1)
	  setDofValues(elIndex, mesh, values, vecs[j]);
      }
    }

    file.close();

    MSG("ARH file read from: %s\n", filename.c_str());
  }


  void ArhReader::setDofValues(int macroElIndex, Mesh *mesh,
			       vector<double>& values, DOFVector<double>* vec)
  {
    FUNCNAME("ArhReader::setDofValues()");

    bool macroElement = false;
    int valuePos = 0;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirstOneMacro(mesh, macroElIndex, -1,
						 Mesh::CALL_EVERY_EL_PREORDER);
    while (elInfo) {
      if (!macroElement) {
	Element *mEl = elInfo->getMacroElement()->getElement();
	for (int i = 0; i < mesh->getGeo(VERTEX); i++)
	  (*vec)[mEl->getDof(i, 0)] = values[valuePos++];	
	macroElement = true;
      }

      Element *el = elInfo->getElement();
      if (!el->isLeaf())
	(*vec)[el->getChild(0)->getDof(mesh->getDim(), 0)] = values[valuePos++];

      elInfo = stack.traverseNext(elInfo);
    }
  }
}
